/*
 *  Copyright (C) 2012 Khaldoon Ghanem
 *  See the file license.txt for copying permission.
 */

#include <stdio.h>
#include <stdlib.h>


#if defined(WIN32)
	#define GLEW_STATIC
	#include <GL/glew.h>
	#include <windows.h>
#else
	#include <GL/glew.h>
#endif

#if defined(__APPLE__) || defined(MACOSX)
	#include <OpenGL/gl.h>
	#include <OpenGL/glu.h>
	#include <GLUT/glut.h>
#else
	#include <GL/glut.h>
	#include <GL/gl.h>
	#include <GL/glu.h>
#endif

#include <string.h>


GLuint v, fm, fn, pm, pn, p;



float w = 400, h = 400;
GLint loc;

float x0 = 0;
float y0 = 0;
float side = 4;
float scale = 1;
float m[9];

void updateView()
{
    float ratio = h / w;

    if (ratio > 1)
        scale = side / w;
    else
        scale = side / h;

    m[0] = scale;
    m[1] = 0;
    m[2] = scale * (-w / 2) + x0;
    m[3] = 0;
    m[4] = scale;
    m[5] = scale * (-h / 2) + y0;
    m[6] = 0;
    m[7] = 0;
    m[8] = 1;

    glUniformMatrix3fv(loc, GL_TRUE, 1, m);
}

void renderScene(void)
{
    updateView();
    glClearColor(0.0, 0.0, 0.0, 0.0);
    glClear (GL_COLOR_BUFFER_BIT);
    glMatrixMode (GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(-1, 1, -1, 1);
    glColor3f(1.0, 1.0, 1.0);
    glBegin (GL_POLYGON);
        glVertex2f(-1, -1);
        glVertex2f(1, -1);
        glVertex2f(1, 1);
        glVertex2f(-1, 1);
    glEnd();
    glutSwapBuffers();
}

int gw, gh;
void changeSize(int ww, int hh)
{
    
    glViewport(0, 0, ww, hh);
    //cx = cx * ww/w;
    //cy = cy * hh/h;
    if (ww != 0)
        w = ww;
    else
        w = 1;
    if (hh != 0)
        h = hh;
    else
        h = 1;
    gh = h;
    gw = w;
}
void changeProgram()
{
    if(p==pm)
    {
        glutSetWindowTitle("Newton Fractals");
        p=pn;
    }
    else
    {
        glutSetWindowTitle("Mandelbrot Set");
        p=pm;
    }
    glUseProgram(p);
    loc = glGetUniformLocation(p, "windowToComplex");
}

int fullscreen = 0;
int screen_pos_x;
int screen_pos_y;
int screen_w;
int screen_h;
void processNormalKeys(unsigned char key, int x, int y)
{

    if (key == 'a')
    {
        float delta = side / 20;
        x0 -= delta;
    }
    if (key == 'd')
    {
        float delta = side / 20;
        x0 += delta;
    }
    if (key == 's')
    {
        float delta = side / 20;
        y0 -= delta;
    }
    if (key == 'w')
    {
        float delta = side / 20;
        y0 += delta;
    }
    if (key == 'q')
    {
        side *= 0.9;
    }
    if (key == 'e')
    {
        side *= 1.1;
    }
    if (key == 'c' || key == 'C')
    {
        x0 = 0;
        y0 = 0;
    }

    if (key == 'r' || key == 'R')
    {
        side = 4;
    }

    if (key == 'p' || key == 'P')
    {
        printf("x0: %f, y0:%lg, side:%f \n", x0, y0, side);
    }
    
    if (key == 'f' || key == 'F')
    {
        fullscreen = !fullscreen;
        if (fullscreen)
        {
            screen_pos_x = glutGet((GLenum) GLUT_WINDOW_X);
            screen_pos_y = glutGet((GLenum) GLUT_WINDOW_Y);
            screen_w = gw;
            screen_h = gh;
            glutFullScreen();
        }
        else
        {
            glutReshapeWindow(screen_w, screen_h);
            glutPositionWindow(screen_pos_x, screen_pos_y);
        }
    }   
    
    if (key == 32)
    {
        changeProgram();
    }
    
    if (key == 27)
        exit(0);
}
float old_x0;
float old_y0;
int pressed = 0;

void mult(float* m, float* x, float* y)
{
    float tx = m[0] * *x + m[1] * *y + m[2];
    float ty = m[3] * *x + m[4] * *y + m[5];
    *x = tx;
    *y = ty;
}

float mouse_x;
float mouse_y;
void mouseButton(int button, int state, int x, int y)
{
    if (button == GLUT_LEFT_BUTTON)
    {

        // when the button is released
        if (state == GLUT_UP)
        {
            pressed = 0;
        }
        else
        { // state = GLUT_DOWN
            mouse_x = x;
            mouse_y = y;
            old_x0 = x0;
            old_y0 = y0;
            pressed = 1;
        }
    }
    // Wheel reports as button 3(scroll up) and button 4(scroll down)
    if ((button == 3 && state == GLUT_DOWN)
            || (button == 4 && state == GLUT_DOWN))
    {
        float zoom;
        if (button == 3)
            zoom = 1.1;
        else
            zoom = 0.9;
        float xx = x;
        float yy = h - y;

        mult(m, &xx, &yy);
        float m2[9];
        m2[0] = zoom;
        m2[1] = 0;
        m2[2] = zoom * (-xx) + xx;
        m2[3] = 0;
        m2[4] = zoom;
        m2[5] = zoom * (-yy) + yy;
        m2[6] = 0;
        m2[7] = 0;
        m2[8] = 1;
        mult(m2, &x0, &y0);
        side = side * zoom;
    }
}

void mouseMove(int x, int y)
{
    if (pressed)
    {
        x0 = old_x0 + (mouse_x - x) * scale;
        y0 = old_y0 - (mouse_y - y) * scale;
    }
}


char *textFileRead(char *fn) {


	FILE *fp;
	char *content = NULL;

	int count=0;

	if (fn != NULL) {
		fp = fopen(fn,"rt");

		if (fp != NULL) {
      
      fseek(fp, 0, SEEK_END);
      count = ftell(fp);
      rewind(fp);

			if (count > 0) {
				content = (char *)malloc(sizeof(char) * (count+1));
				count = fread(content,sizeof(char),count,fp);
				content[count] = '\0';
			}
			fclose(fp);
		}
	}
	return content;
}


#define printOpenGLError() printOglError(__FILE__, __LINE__)

int printOglError(char *file, int line)
{
    //
    // Returns 1 if an OpenGL error occurred, 0 otherwise.
    //
    GLenum glErr;
    int retCode = 0;

    glErr = glGetError();
    while (glErr != GL_NO_ERROR)
    {
        printf("glError in file %s @ line %d: %s\n", file, line,
                gluErrorString(glErr));
        retCode = 1;
        glErr = glGetError();
    }
    return retCode;
}

void printShaderInfoLog(GLuint obj)
{
    int infologLength = 0;
    int charsWritten = 0;
    char *infoLog;

    glGetShaderiv(obj, GL_INFO_LOG_LENGTH, &infologLength);

    if (infologLength > 0)
    {
        infoLog = (char *) malloc(infologLength);
        glGetShaderInfoLog(obj, infologLength, &charsWritten, infoLog);
        if (strlen(infoLog) > 0)
            printf("%s\n", infoLog);
        free(infoLog);
    }
}

void printProgramInfoLog(GLuint obj)
{
    int infologLength = 0;
    int charsWritten = 0;
    char *infoLog;

    glGetProgramiv(obj, GL_INFO_LOG_LENGTH, &infologLength);

    if (infologLength > 0)
    {
        infoLog = (char *) malloc(infologLength);
        glGetProgramInfoLog(obj, infologLength, &charsWritten, infoLog);
        if (strlen(infoLog) > 0)
            printf("%s\n", infoLog);
        free(infoLog);
    }
}


void setShaders()
{

    char *vs = NULL;
    char *fms = NULL;
    char *fns = NULL;

    v = glCreateShader(GL_VERTEX_SHADER);
    fm = glCreateShader(GL_FRAGMENT_SHADER);
    fn = glCreateShader(GL_FRAGMENT_SHADER);

    vs = textFileRead("fixed.vert");
    fms = textFileRead("mandel.frag");
    fns = textFileRead("newton.frag");

    const char * vv = vs;
    const char * ffm = fms;
    const char * ffn = fns;    

    glShaderSource(v, 1, &vv, NULL);
    glShaderSource(fm, 1, &ffm, NULL);
    glShaderSource(fn, 1, &ffn, NULL);    

    free(vs);
    free(fms);
    free(fns);    

    glCompileShader(v);
    int isCompiled;
    glGetShaderiv(v, GL_COMPILE_STATUS, &isCompiled);
    if(isCompiled == 0)
    {
        printf("Vertex Shader - Compilation Error!\n");
        printShaderInfoLog(fm);
        exit(1);
    }
    
    
    glCompileShader(fm);
    glGetShaderiv(fm, GL_COMPILE_STATUS, &isCompiled);
    if(isCompiled == 0)
    {
        printf("Mandel Fragment Shader - Compilation Error!\n");
        printShaderInfoLog(fm);
        exit(1);
    }
    
    
    glCompileShader(fn);    
    glGetShaderiv(fn, GL_COMPILE_STATUS, &isCompiled);
    if(isCompiled == 0)
    {
        printf("Newton Fragment Shader - Compilation Error!\n");
        printShaderInfoLog(fn);
        exit(1);
    }

    //////////////////////////////////////////////////////////////////
    pm = glCreateProgram();
    glAttachShader(pm, v);
    glAttachShader(pm, fm);  

    glLinkProgram(pm);
    int isLinked;
    glGetProgramiv(pm, GL_LINK_STATUS, (int *)&isLinked);
    if(isLinked == 0)
    {
        printf("Mandelbrot Program - Linking Error!\n");
        printProgramInfoLog(pm);
        exit(1);
    }


    pn = glCreateProgram();
    glAttachShader(pn, v);
    glAttachShader(pn, fn);    

    glLinkProgram(pn);
    glGetProgramiv(pn, GL_LINK_STATUS, (int *)&isLinked);
    if(isLinked == 0)
    {
        printf("Newton Program - Linking Error!\n");
        printProgramInfoLog(pn);
        exit(1);
    }
    p = pn;
    changeProgram();
}

int main(int argc, char **argv)
{
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA);
    glutInitWindowPosition(100, 100);
    glutInitWindowSize(w, h);
    glutCreateWindow("If you see this, then something went wrong!");

    glutDisplayFunc(renderScene);
    glutIdleFunc(renderScene);    
    glutReshapeFunc(changeSize);
    glutKeyboardFunc(processNormalKeys);
    glutMouseFunc(mouseButton);
    glutMotionFunc(mouseMove);

    glClearColor(1.0, 1.0, 1.0, 1.0);

    glewInit();
    if (!glewIsSupported("GL_VERSION_2_0"))
    {
        printf("OpenGL 2.0 not supported\n");
        exit(1);
    }

    setShaders();

    char* usageInfo = textFileRead("usageInfo.txt");
    printf("%s", usageInfo);
    free(usageInfo);

    glutMainLoop();

    return 0;
}

