//Newton Fractals
/*
 *  Copyright (C) 2012 Khaldoon Ghanem
 *  See the file license.txt for copying permission.
 */

#version 120
uniform mat3 windowToComplex;

//Playing with these variables gives different brightness patterns
int maxIter = 100;
float converg = 0.1; //speed of convergence
float eps = 0.1; //radius of circle around each root; within this circle, we have convereged to the corresponding root

vec2 root1 = vec2(1.0, 0);
vec2 root2 = vec2(-0.5, sqrt(3.0)/2.0);
vec2 root3 = vec2(-0.5, -sqrt(3.0)/2.0);
vec4 hue = vec4(0.0, 0.0, 0.0,1.0);

float cAbsSq(vec2 z)
{
	return z.x*z.x+z.y*z.y;
}

vec2 cMul(vec2 z1, vec2 z2)
{
	return vec2(z1.x*z2.x-z1.y*z2.y, z1.x*z2.y+z1.y*z2.x);
}

vec2 cDiv(vec2 z1, vec2 z2)
{
	float r = z2.x*z2.x+z2.y*z2.y;
	return vec2((z1.x*z2.x+z1.y*z2.y)/r,(z1.y*z2.x-z1.x*z2.y)/r);
}

void main()
{
	
	vec2 z = vec2(windowToComplex * vec3(gl_FragCoord.x, gl_FragCoord.y, 1));

	int iter = 0;
	
	for(;iter<maxIter;iter++)
	{
		vec2 z2 = cMul(z,z);
		vec2 z3 = cMul(z2,z);
		vec2 temp = z - converg* cDiv(z3- vec2(1.0, 0), 3.0* z2);
		z = temp;
		if(cAbsSq(z - root1)<eps)
		{
			hue = vec4(1.0, 0, 0,1.0);
			break;
		}
		if(cAbsSq(z - root2)<eps)
		{
			hue = vec4(0, 1.0, 0,1.0);			
			break;
		}
		if(cAbsSq(z - root3)<eps)
		{
			hue = vec4(0, 0, 1.0,1.0);			
			break;
		}				
	}

	//dis = dis * iter;
	float intensity = 1.0-float(iter)/float(maxIter);

	gl_FragColor = intensity * hue;

}


